# -*- coding: utf-8 -*-

import inspect
import logging
import sys

from argparse import ArgumentParser, Action as ArgumentAction, Namespace
from itertools import product
from pathlib import Path
from types import GeneratorType
from typing import Union, Iterable, Mapping, List, MutableMapping

from allpairspy import AllPairs
from colorlog import ColoredFormatter
from tabulate import tabulate

import matrix_runner.preferences as prefs

from .action import Action
from .axis import Axis
from .command import Result
from .config import Config
from .filter import Filter
from ._helper import join, reduce_dict, log_formatter, colorize


class _AppendAxisValuesAction(ArgumentAction):
    def __call__(self, parser, namespace, values, option_string=None):
        items = getattr(namespace, self.dest, None)
        if items is None:
            items = []
        if isinstance(values, Iterable):
            items.extend(values)
        else:
            items.append(values)
        setattr(namespace, self.dest, items)


class LogFormatter(ColoredFormatter):
    """Custom log formatter"""

    class Record:
        """Custom log record carrying additional 'cmd' attribute."""

        def __init__(self, record):
            """Add attributes from the escape_codes dict and the record."""
            self.__dict__.update({'cmd': ''})
            self.__dict__.update(record.__dict__)

            # Keep a reference to the original record so ``__getattr__`` can
            # access functions that are not in ``__dict__``
            self.__record = record

        def __getattr__(self, name):
            return getattr(self.__record, name)

    def format(self, record):
        record = self.Record(record)
        return super().format(record)


class Runner:
    """The matrix Runner base class.
    """

    def __init__(self):
        self._axes: MutableMapping[str, Axis] = {}
        self._actions: MutableMapping[str, Action] = {}
        self._filter = []
        self._args = None
        self._matrix = None
        self._records = {}

    @property
    def axes(self) -> Mapping[str, Axis]:
        """Dictionary of matrix configuration Axes used by this Runner."""
        return self._axes

    @property
    def actions(self) -> Mapping[str, Action]:
        """Dictionary of Actions provided by this Runner."""
        return self._actions

    def add_axis(self, axes: Union[Axis, Iterable[Axis]]) -> None:
        """Add Axes to the Runner matrix.

        Args:
            axes: Axis/axes to be added.

        Raises:
            TypeError if axis argument has inappropriate type.
            ValueError if an Axis with same name is already part of the matrix.
        """
        if isinstance(axes, GeneratorType):
            axes = list(axes)
        if isinstance(axes, Iterable) and all(isinstance(a, Axis) for a in axes):
            for axis in axes:
                self.add_axis(axis)
        elif isinstance(axes, Axis):
            if axes.name in self._axes:
                raise ValueError(f'Runner matrix already contains an Axis with name {axes.name}!')
            self._axes[axes.name] = axes
        else:
            raise TypeError(f'Parameter axis must be of type Axis not {str(type(axes))}!')

    def add_action(self, actions) -> None:
        """Add Action to the Runner

        Args:
            actions: Action(s) to be added

        Raises:
            TypeError if action argument is not of type Action.
            ValueError if action with same name already part of the configuration.
        """
        if isinstance(actions, GeneratorType):
            actions = list(actions)
        if isinstance(actions, Iterable) and all(isinstance(a, Action) for a in actions):
            for action in actions:
                self.add_action(action)
        elif isinstance(actions, Action):
            if actions.name in self._actions:
                raise ValueError(f'Runner matrix already contains an Action with name {actions.name}!')
            self._actions[actions.name] = actions
        else:
            raise TypeError(f'Parameter action must be of type Action not {str(type(actions))}!')

    @property
    def _arg_parser(self):
        """The ArgumentParser to run matrix configurations."""
        parser = ArgumentParser()
        group = parser.add_mutually_exclusive_group()
        group.add_argument("--silent", action='store_true', help="Silent mode, only errors are shown.")
        group.add_argument("--verbose", action='store_true', help="Verbose log output.")
        group.add_argument("--debug", action='store_true', help="Debug log output.")
        parser.add_argument("--pairwise", "-2", action='store_true',
                            help="Reduce number of combinations using pairwise algorithm.")
        for axis in self._axes.values():
            flags = ["--" + axis.name]
            if axis.abbrev:
                flags += ["-" + axis.abbrev]
            parser.add_argument(*flags, type=axis.converter(), action=_AppendAxisValuesAction,
                                help=f"{axis.desc}: {', '.join([join(v) for v in axis])}")
        action_help = f"Action(s) to be executed: {', '.join(self._actions.keys())}"
        parser.add_argument("action", choices=list(self._actions.values()) + [[]], nargs='*',
                            metavar="action", type=self._actions.get, help=action_help)
        return parser

    def _parse_args(self, argv: List[str] = None) -> Namespace:
        """Parse the arguments using the ArgumentParser.
        The list of values for omitted matrix axes defaults to all possible values.

        Args:
            argv: List of command line arguments

        Returns:
            Namespace with all command line flags and arguments.
        """
        args = self._arg_parser.parse_args(argv)
        for axis in self._axes.values():
            if getattr(args, axis.name, None) is None:
                setattr(args, axis.name, axis.values)
        return args

    def run(self, argv: List[str] = None):
        """Run with given command line arguments."""
        args = self._parse_args(argv)

        if args.silent:
            logging.root.setLevel(logging.ERROR)
        if args.verbose:
            logging.root.setLevel(logging.INFO)
        if args.debug:
            logging.root.setLevel(logging.DEBUG)

        axes = []
        for axis in self._axes:
            axes.append([(axis, v) for v in getattr(args, axis)])
        if args.pairwise:
            self._matrix = [Config(**dict(m)) for m in
                            AllPairs(axes, filter_func=lambda row: not Filter.match(Config(**dict(row))))]
        else:
            self._matrix = [Config(**dict(m)) for m in product(*axes) if not Filter.match(Config(**dict(m)))]
        for config in self._matrix:
            self.run_config(args.action, config)
        return all(r.success for r in reduce_dict(self._records))

    def run_config(self, actions: List[Action], config: Config):
        """Run all actions for the given configuration."""
        for action in actions:
            cfg = "][".join([self._axes[k].tostring(vars(config)[k]) for k in sorted(self._axes.keys())])
            fmt = LogFormatter(f"%({prefs.prefix_colors()['config']})s[{cfg}]"
                               f"%({prefs.prefix_colors()['action']})s({action.name}"
                               f"%(cmd)s)%(reset)s %(log_color)s%(message)s")
            with log_formatter(fmt):
                results = action(config)
                self.record_results(action, config, results)

    def record_results(self, action: Action, config: Config, results: List[Result]):
        """Record execution results."""
        self._records_for(config)[action.name] = results

    def records_for(self, config: Config) -> Mapping[str, List[Result]]:
        """Retrieve all execution records for given config."""
        return dict(self._records_for(config))

    def _records_for(self, config: Config) -> MutableMapping[str, List[Result]]:
        config_dict = vars(config)
        records = self._records
        for k in sorted(config_dict.keys()):
            records = records.setdefault(config_dict[k], {})
        return records


class RunnerApplication(Runner):
    """Runner"""

    def __init__(self, argv=None, axes=None, actions=None):
        super().__init__()

        self._configure_logging()

        logging.root.setLevel(prefs.log_level())
        logging.root.updateColors(prefs.log_colors())  # pylint: disable=no-member

        script_dir = Path(inspect.stack()[-1][1]).resolve().parent
        work_dir = Path.cwd()
        if script_dir != work_dir:
            logging.error("====[ WARNING ]====")
            logging.error("The default Matrix Runner application expects the "
                          "script to be located in the current working "
                          "directory!")
            logging.error(colorize(f"Expected working directory: %(blue)s{script_dir}"))
            logging.error(colorize(f"Actual working directory: %(blue)s{work_dir}"))
            logging.error("===================")

        if axes is None:
            axes = Axis.get_instances()
        if actions is None:
            actions = Action.get_instances()
        self.add_axis(axes)
        self.add_action(actions)

        success = self.run(argv)
        self.summary()

        if not success:
            sys.exit(1)
        sys.exit(0)

    # noinspection SpellCheckingInspection
    @staticmethod
    def _configure_logging(level=logging.WARNING):
        fmt = ColoredFormatter("%(log_color)s%(message)s")

        stdouth = logging.StreamHandler(sys.stdout)
        stdouth.addFilter(lambda record: record.levelno <= logging.WARNING)
        stdouth.setFormatter(fmt)
        logging.root.addHandler(stdouth)

        stderrh = logging.StreamHandler(sys.stderr)
        stderrh.setFormatter(fmt)
        errfn = stderrh.setLevel
        stderrh.setLevel = lambda ll: errfn(ll) if ll >= logging.ERROR else errfn(logging.ERROR)
        logging.root.addHandler(stderrh)

        fn = logging.root.setLevel
        logging.root.setLevel = lambda l: (fn(l), stdouth.setLevel(l), stderrh.setLevel(l))
        logging.root.setLevel(level)

        type(logging.root).updateColors = lambda s, c: fmt.log_colors.update(c)

    def summary(self):
        """Write full execution summary to log."""
        headers = [colorize(f"%({prefs.prefix_colors()['config']})s{a}") for a in sorted(self._axes.keys())]
        headers += [colorize(f"%({prefs.prefix_colors()['action']})s{a}") for a in sorted(self._actions.keys())]
        rows = []
        for config in self._matrix:
            records = self.records_for(config)
            row = [
                colorize(f"%({prefs.summary_colors()['config']})s{self._axes[k].tostring(vars(config)[k])}")
                for k in sorted(self._axes.keys())
            ]
            row.extend([self._actions[k].get_summary(records.get(k)) for k in sorted(self._actions.keys())])
            rows.append(row)

        logging.warning("")
        logging.warning("Matrix Summary")
        logging.warning("==============")
        logging.warning("")
        logging.warning(tabulate(rows, headers=headers))



# -*- coding: utf-8 -*-

import logging
import weakref

from functools import reduce
from inspect import signature
from typing import Callable, List, Iterator, Tuple, Union, AnyStr, Optional

import matrix_runner.preferences as prefs

from ._helper import colorize
from .command import Command, Result
from .config import Config
from .report import ReportFilter

ActionFunction = Callable[[Config, Optional[Result]], Iterator[Union[Command, Tuple]]]


class Action:
    """Wrapper class for a matrix runner actions function.
    """

    _instances = set()

    def __init__(self, name: AnyStr, fn: ActionFunction, desc: AnyStr = None):
        """Initialize a new runner Action.

        Args:
            name: The unique name to identify this Action.
            fn: The function
            desc: Descriptive text to document the intention of this Action.
        """
        self._name = name
        self._fn = fn
        self._get_summary = None
        self._desc = desc
        self._instances.add(weakref.ref(self))

    def __del__(self):
        self._instances.discard(weakref.ref(self))

    def __str__(self):
        return self._name

    def __repr__(self):
        return self._name

    def __call__(self, config: Config) -> List[Result]:
        results = []
        params = [config]
        if len(signature(self._fn).parameters) >= 2:
            params += [results]
        cmds = self._fn(*params)
        try:
            for cmd in cmds:
                results.append(cmd())
        except RuntimeError as e:
            logging.error(e)
            logging.debug(e, exc_info=True)
            results.append(Result())
        return results

    def summary(self, fn: Callable[[List[Result]], str]) -> Callable[[List[Result]], str]:
        """Record custom summary function."""
        self._get_summary = fn
        return fn

    @classmethod
    def get_instances(cls) -> Iterator:
        """Retrieve all class instances."""
        dead = set()
        for ref in cls._instances:
            obj = ref()
            if obj is not None:
                yield obj
            else:
                dead.add(ref)
        cls._instances -= dead

    @property
    def name(self) -> str:
        """The unique name to identify this Action."""
        return self._name

    @property
    def desc(self) -> str:
        """Descriptive text to document the intention of this Action."""
        return self._desc

    def get_summary(self, results: List[Result]) -> str:
        """Create a summary for this action."""
        if not results:
            return colorize(f"%({prefs.summary_colors()['skip']})s(skip)")
        try:
            if all(r.success for r in results):
                if self._get_summary:
                    return colorize(self._get_summary(results))
                if any(isinstance(r.test_report, ReportFilter.Summary) for r in results):
                    results_with_reports = filter(lambda r: isinstance(r.test_report, ReportFilter.Summary), results)
                    report_summaries = map(lambda r: r.test_report.summary, results_with_reports)
                    passed, executed = reduce(lambda a, b: (a[0]+b[0], a[1]+b[1]), report_summaries, (0, 0))
                    if passed == executed:
                        color = prefs.summary_colors()['success']
                    else:
                        color = prefs.summary_colors()['unstable']
                    return colorize(f"%({color})s{passed}/{executed}")
                return colorize(f"%({prefs.summary_colors()['success']})ssuccess")
        except RuntimeError as e:
            logging.debug(e, exc_info=True)
            return colorize(f"%({prefs.summary_colors()['error']})sERROR")
        return colorize(f"%({prefs.summary_colors()['fail']})sFAILED")



# -*- coding: utf-8 -*-

import platform
import os

from argparse import ArgumentParser
from pathlib import Path
from shutil import copyfile

from .inspect import InspectRunner
from .preferences import Preferences


def inspect(_, argv):
    """Inspect script

    Args:
        _: (ignored)
        argv (list): Remaining command line arguments to be passed on
    """
    InspectRunner(argv)


def config(args, _):
    """Configure preferences

    Args:
        args (Namespace): Arguments from the parser
        _: (ignored)
    """
    if args.system:
        name = "system"
        conf = Preferences.system_conf
    if args.user:
        name = "user"
        conf = Preferences.user_conf
    if args.local:
        name = "local"
        conf = Path.cwd().joinpath(Preferences.local_conf.name)
    if not args.action:
        print(f"The {name} config file is located at:")
        print(conf)
    if args.action == "list":
        if conf.exists():
            with open(conf, encoding='utf-8') as file:
                print(file.read())
        else:
            print(f"No {name} config file exists at:")
            print(conf)
    if args.action == "init":
        copyfile(Preferences.system_conf, conf)
    if args.action == "edit":
        if platform.system() == "Windows":
            os.startfile(conf)  # pylint: disable=no-member
        else:
            os.system(f"editor {conf}")


def main():
    """Main function"""
    parser = ArgumentParser(prog='python -m matrix_runner')
    subparsers = parser.add_subparsers(dest='cmd', help='sub-command help')

    config_parser = subparsers.add_parser('config', help='Configure matrix-runner preferences')
    group = config_parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-s', '--system', action='store_true', help="System defaults.")
    group.add_argument('-u', '--user', action='store_true', help="User settings.")
    group.add_argument('-l', '--local', action='store_true', help="Local folder settings.")
    config_parser.add_argument('action', choices=['list', 'init', 'edit'],  nargs='?', help="Action")

    subparsers.add_parser('inspect', help='Inspect a matrix build script')

    args, argv = parser.parse_known_args()

    __import__(__name__).__dict__[args.cmd](args, argv)


main()



# -*- coding: utf-8 -*-
import io
import logging

from contextlib import contextmanager
from enum import Enum
from fnmatch import fnmatch
from inspect import signature, Parameter
from typing import Tuple, Union, Any, AnyStr
from functools import reduce, partial


def colorize(msg):
    """Colorize a message string with ANSI colors.

    Args:
        msg: The message string containing color codes.

    Returns:
        Message with replaced color codes.
    """
    from colorlog.escape_codes import escape_codes  # pylint: disable=import-outside-toplevel
    msg += "%(reset)s"
    return msg % escape_codes


@contextmanager
def log_formatter(fmt):
    """Context manager for root log formatters.

    The formatter of all root handlers is temporarily
    replaced with the given one.

    Args:
         fmt: Log formatter to be used, temporarily.
    """
    old_formatter = {}
    for hnd in logging.root.handlers:
        old_formatter[hnd] = hnd.formatter
        hnd.formatter = fmt
    yield
    for hnd in logging.root.handlers:
        hnd.formatter = old_formatter[hnd]


def fnmatch_ex(value: Union[str, Tuple, Enum], pattern: str) -> bool:
    """Extended version of function fnmatch.fnmatch accepting Axis values.

    The value can one of
        str: The call is forwarded to fnmatch.
        Enum: The value associated with the enumerate is processed, recursively.
        tuple: fnmatch is called for all tuple members, any match is accepted.

    Args:
        value: Axis value to be matched against.
        pattern: fnmatch pattern to match.

    Returns:
        True if pattern matches value, False otherwise.

    Raises:
        TypeError if value is of unsupported type.
    """
    if isinstance(value, str):
        return fnmatch(value, pattern)
    if isinstance(value, Enum):
        return fnmatch_ex(value.value, pattern)
    if isinstance(value, tuple):
        return any(fnmatch_ex(v, pattern) for v in value)
    return fnmatch(str(value), pattern)


def join(value, delim: str = "|"):
    """Flatten an arbitrary Axis value to a simple string representation.

    For Enum values flatten is called recursively with the associated value.
    For tuples flatten is called recursively for all contained values. The
    resulting list is joined with ch.
    String values are returned unchanged.
    All other types are converted to String.

    Args:
        value: Value converted to string.
        delim: Character to be used to join collections.

    Returns:
        String representation of input value, tuples joined with ch.
    """
    if isinstance(value, Enum):
        return join(value.value, delim)
    if isinstance(value, tuple):
        return delim.join([join(v, delim) for v in value])
    if isinstance(value, str):
        return value
    return str(value)


def reduce_dict(value):
    """Reduce dict values, recursively.

    Args:
        value: Any value.

    Returns:
        Reduced dict or plain value.
    """
    if isinstance(value, dict):
        return reduce(lambda x, y: x + reduce_dict(y), value.values(), [])
    return value


def is_msys_cygwin_tty(stream):
    """Check if running on a MSYS/Cygwin TTY"""
    # https://github.com/tartley/colorama/pull/226
    # pylint: disable=import-outside-toplevel, invalid-name, missing-class-docstring, too-few-public-methods
    try:
        import msvcrt
        import ctypes
        import re
    except ImportError:
        return False

    if not hasattr(stream, "fileno"):
        return False

    if not hasattr(ctypes, "windll") or not hasattr(ctypes.windll.kernel32, "GetFileInformationByHandleEx"):
        return False

    try:
        fileno = stream.fileno()
    except io.UnsupportedOperation:
        return False
    handle = msvcrt.get_osfhandle(fileno)
    FileNameInfo = 2

    class FILE_NAME_INFO(ctypes.Structure):
        _fields_ = [('FileNameLength', ctypes.c_ulong),
                    ('FileName', ctypes.c_wchar * 40)]

    info = FILE_NAME_INFO()
    ret = ctypes.windll.kernel32.GetFileInformationByHandleEx(handle,
                                                              FileNameInfo,
                                                              ctypes.byref(info),
                                                              ctypes.sizeof(info))
    if ret == 0:
        return False

    msys_pattern = r"\\msys-[0-9a-f]{16}-pty\d-(to|from)-master"
    cygwin_pattern = r"\\cygwin-[0-9a-f]{16}-pty\d-(to|from)-master"
    return re.match(msys_pattern, info.FileName) is not None or \
           re.match(cygwin_pattern, info.FileName) is not None



class partial_with_inspect(partial):  # pylint: disable=invalid-name
    """Function object to bind a function to arguments.

    This subclass of functools.partial enhances the function object
    with inspect functionality.
    """

    def __new__(cls, func, /, *args, **keywords):
        self = super().__new__(cls, func, *args, **keywords)
        self._arg_cache = None
        return self

    @property
    def __name__(self):
        """The name of the wrapped function."""
        return self.func.__name__

    def __getattr__(self, item: AnyStr):
        """Dynamically provide named access to bound function arguments.

        Args:
            item: An argument name.

        Returns:
            Value bound to the argument.
        """
        if item in self.keywords:
            return self.keywords[item]

        index, kind = self._arg_index(item)
        if index is not None:
            if kind == Parameter.VAR_POSITIONAL:
                return self.args[index:]
            if kind == Parameter.VAR_KEYWORD:
                return self.keywords
            return self.args[index]
        raise AttributeError(f"'partial_with_inspect<{self.__name__}>' object has no attribute '{item}'")

    def _arg_index(self, item: AnyStr) -> Tuple[int, Any]:
        # pylint: disable=access-member-before-definition, attribute-defined-outside-init
        if not self._arg_cache:
            sig = signature(self.func)
            self._arg_cache = {k: (pos, sig.parameters[k].kind) for pos, k in enumerate(sig.parameters)}
        return self._arg_cache.get(item, (None, None))



# -*- coding: utf-8 -*-

import inspect
import logging

from configparser import ConfigParser
from pathlib import Path
from site import USER_BASE

from colorlog import default_log_colors


__all__ = ['Preferences', 'log_level', 'log_colors', 'prefix_colors', 'summary_colors']


class Preferences:
    """Preferences capsule"""

    system_conf = Path(__file__).parent.joinpath("default.conf")
    user_conf = Path(USER_BASE).joinpath("matrix_runner.conf")
    local_conf = Path(inspect.stack()[-1][1]).resolve().parent.joinpath("matrix_runner.conf")
    instance = None

    def __init__(self, system: Path = system_conf, user: Path = user_conf, local: Path = local_conf):
        """Load preferences from multi-leveled configuration files

        Local preferences take precedence over user and system levels.
        User preferences take precedence over system level.
        System preferences take precedence over built-in fallback values.

        Args:
            system: Path to the system level configuration file
            user: Path to the user level configuration file
            local: Path to the local configuration file
        """
        self._conf = ConfigParser()
        self._conf.read(system)
        self._conf.read(user)
        self._conf.read(local)

    @property
    def log_level(self):
        """Default log level"""
        level = self._conf.get('global', 'loglevel', fallback='WARNING')
        levelnum = logging.getLevelName(level)
        if isinstance(levelnum, int):
            return levelnum
        logging.warning("Invalid log level defined in config file: '%s'!", level)
        return logging.WARNING

    @property
    def log_colors(self):
        """Colors to be used for log output."""
        return {k: self._conf.get('colors', k, fallback=v) for k, v in default_log_colors.items()}

    @property
    def prefix_colors(self):
        """Colors to be used for log prefix."""
        return {
            'config': self._conf.get('colors', 'config', fallback='blue'),
            'action': self._conf.get('colors', 'action', fallback='purple'),
        }

    @property
    def summary_colors(self):
        """Colors to be used for summaries."""
        return {
            'config': self._conf.get('colors', 'summary_config', fallback='bold_blue'),
            'success': self._conf.get('colors', 'summary_success', fallback='green'),
            'unstable': self._conf.get('colors', 'summary_unstable', fallback='yellow'),
            'skip': self._conf.get('colors', 'summary_skip', fallback='yellow'),
            'fail': self._conf.get('colors', 'summary_fail', fallback='bold_red'),
            'error': self._conf.get('colors', 'summary_error', fallback='bold_red')
        }


def log_level():
    """Default log level"""
    if not Preferences.instance:
        Preferences.instance = Preferences()
    return Preferences.instance.log_level


def log_colors():
    """Colors to be used for log output."""
    if not Preferences.instance:
        Preferences.instance = Preferences()
    return Preferences.instance.log_colors


def prefix_colors():
    """Colors to be used for log prefix."""
    if not Preferences.instance:
        Preferences.instance = Preferences()
    return Preferences.instance.prefix_colors


def summary_colors():
    """Colors to be used for summaries."""
    if not Preferences.instance:
        Preferences.instance = Preferences()
    return Preferences.instance.summary_colors



# -*- coding: utf-8 -*-

import logging
import shutil

from contextlib import contextmanager
from io import StringIO
from pathlib import Path
from subprocess import Popen, PIPE, TimeoutExpired
from tempfile import gettempdir
from threading import Thread
from time import perf_counter, time, sleep
from typing import Callable, List, Union, AnyStr, Optional, IO, Iterable, Dict

import psutil

from filelock import FileLock
from psutil import AccessDenied, NoSuchProcess, TimeoutExpired as PSTimeoutExpired

from ._helper import partial_with_inspect
from .report import ReportFilter

CommandFunction = Callable[..., List]


class Result:
    """Execution result capsule."""

    def __init__(self, extra=None):
        self.exit_status = None
        self.success = False
        self.output = StringIO()
        self.test_report = None
        self.start_time = 0
        self.end_time = 0
        self.start_perf_counter = 0
        self.end_perf_counter = 0
        self._extra = extra

    @contextmanager
    def capture(self, proc: Popen):
        """Context manager for capturing process output."""
        output_handler = logging.StreamHandler(self.output)
        logger = logging.getLogger(f'{self.__class__.__name__}:{proc.pid}')
        logger.addHandler(output_handler)
        logger.setLevel(logging.INFO)

        def capture_output(stream: IO, loglevel: int):
            for line in stream:
                line = line.rstrip()
                if line:
                    logger.log(loglevel, line, extra=self._extra)

        thout = Thread(target=capture_output, args=(proc.stdout, logging.INFO), daemon=True)
        therr = Thread(target=capture_output, args=(proc.stderr, logging.ERROR), daemon=True)
        thout.start()
        therr.start()

        self.log_start()
        try:
            yield self
        finally:
            self.log_end()

            thout.join()
            therr.join()

    def log_start(self):
        """Record execution start time stamp."""
        self.start_time = time()
        self.start_perf_counter = perf_counter()

    def log_end(self):
        """Record execution end time stamp."""
        self.end_time = time()
        self.end_perf_counter = perf_counter()


class Command:
    """Command wrapper"""
    _exit_code: Union[int, Callable[[int], bool]]

    def __init__(self, fn: Callable[[], List], exit_code: Union[int, Iterable[int], Callable[[int], bool]] = 0,
                 needs_shell: bool = False, encoding: str = 'utf-8', exclusive: bool = False,
                 timeout: Optional[float] = None, retry: Optional[int] = 1, rest_period: Optional[float] = None,
                 test_report: Optional[ReportFilter] = None):  # pylint: disable=too-many-arguments
        self._fn = fn
        self._exit_code = exit_code
        self._needs_shell = needs_shell
        self._encoding = encoding
        self._exclusive = exclusive
        self._timeout = timeout
        self._retry = max(retry, 1)
        self._rest_period = rest_period
        self._test_report = test_report
        self.extra = {'cmd':  f":{fn.__name__}"}

    def __call__(self):
        cmdline = self._fn()
        result = CommandResult(self, cmdline)
        self._execute(cmdline, result)
        return result

    def __getattr__(self, item):
        if isinstance(self._fn, partial_with_inspect):
            return self._fn.__getattr__(item)
        raise AttributeError(f"'Command' object has no attribute '{item}'")

    @staticmethod
    def _resolve_cmdlineel(elem) -> str:
        """Resolve command line element."""
        if isinstance(elem, str):
            return elem
        if isinstance(elem, Path):
            return str(elem.resolve())
        return str(elem)

    def _resolve_cmdline(self, cmdline: List) -> List[str]:
        """Resolve command line."""
        cmdline = [Command._resolve_cmdlineel(el) for el in cmdline]
        cmd = shutil.which(cmdline[0])
        if cmd:
            if ' ' in cmd:
                cmd = f'"{cmd}"'
            cmdline[0] = cmd
        else:
            logging.error("Command '%s' not found in current environment.", cmdline[0], extra=self.extra)
        return cmdline

    # pylint: disable=too-many-arguments,too-many-statements
    @staticmethod
    def _popen(cmdline: List[str], result: Result, needs_shell: bool = False, encoding: str = 'utf8',
               timeout: Optional[float] = None, extra: Optional[Dict[str, str]] = None) -> int:

        @contextmanager
        def observe(process):
            parent = psutil.Process(process.pid)
            children = {parent}

            def observer():
                try:
                    logging.debug("Observing process %s [pid %d]", parent.name(), parent.pid, extra=extra)
                    while parent.is_running():
                        sleep(1)
                        try:
                            new = set(parent.children(recursive=True)) - children
                            children.update(new)
                            for child in new:
                                logging.debug("Detected new subprocess %s [pid %d]",
                                              child.name(), child.pid, extra=extra)
                        except NoSuchProcess:
                            pass
                except NoSuchProcess:
                    pass

            worker = Thread(target=observer, daemon=True)
            worker.start()

            try:
                yield
            finally:
                if parent.is_running():
                    try:
                        logging.error("Process %s [pid %d] did not terminate!", parent.name(), parent.pid, extra=extra)
                        parent.kill()
                        parent.wait(timeout=10)
                    except NoSuchProcess:
                        pass
                    except PSTimeoutExpired:
                        logging.critical("Unable to kill process %s [pid %d]!", parent.name(), parent.pid, extra=extra)
                    # Workaround for zombie processes
                    # Zombie processes are assumed as still running,
                    # hence the worker (observer) doesn't terminate.
                    parent._gone = True  # pylint: disable=protected-access
                worker.join()

                logging.debug("Waiting for all subprocesses to terminate...", extra=extra)
                try:
                    _, alive = psutil.wait_procs(children, timeout=1)
                    for pid in alive:
                        logging.error("Subprocess %s [pid %d] did not terminate!", pid.name(), pid.pid, extra=extra)
                        try:
                            pid.kill()
                            pid.wait(timeout=10)
                        except NoSuchProcess:
                            pass
                        except PSTimeoutExpired:
                            logging.critical("Unable to kill subprocess %s [pid %d]!", pid.name(), pid.pid, extra=extra)
                except NoSuchProcess:
                    pass
                except AccessDenied as exc_info:
                    logging.warning("Waiting for subprocesses failed!", extra=extra)
                    logging.debug("Exception Info", exc_info=exc_info, extra=extra)

        if isinstance(cmdline, Iterable) and needs_shell:
            cmdline = ' '.join(cmdline)

        with Popen(cmdline, stdout=PIPE, stderr=PIPE, shell=needs_shell, encoding=encoding) as proc, \
                result.capture(proc), observe(proc):
            try:
                result.exit_status = proc.wait(timeout)
            except TimeoutExpired as e:
                logging.debug("Command timeout expired, terminating...")
                proc.terminate()
                result.exit_status = TimeoutError(e)
                raise TimeoutError() from e

        return result.exit_status

    def _is_success(self, rc: int) -> bool:
        if isinstance(self._exit_code, Callable):
            return self._exit_code(rc)
        if isinstance(self._exit_code, Iterable):
            return rc in self._exit_code
        return rc == self._exit_code

    def _execute_with_timeout(self, cmdline: List[AnyStr], result: Result) -> bool:
        try:
            if self._timeout:
                logging.info("... with timeout of %d seconds", self._timeout, extra=self.extra)

            rc = Command._popen(cmdline, result, needs_shell=self._needs_shell, encoding=self._encoding,
                                timeout=self._timeout, extra=self.extra)
            result.success = self._is_success(rc)

            if result.success:
                logging.warning("%s succeeded with exit code %d", cmdline[0], rc, extra=self.extra)
            else:
                logging.error("%s failed with exit code %d", cmdline[0], rc, extra=self.extra)

        except TimeoutError:
            logging.error("Execution timed out!", extra=self.extra)

        return result.success

    def _execute_with_rest(self, cmdline: List[AnyStr], result: Result) -> bool:
        if self._rest_period:
            logging.info("... resting for %d seconds", self._rest_period, extra=self.extra)
            sleep(self._rest_period)

        return self._execute_with_timeout(cmdline, result)

    def _execute_with_retry(self, cmdline: List[AnyStr], result: Result) -> bool:
        tries = self._retry
        for i in range(0, tries):
            logging.log(logging.DEBUG if not i else logging.WARN, "... with retry %d/%d", i+1, tries, extra=self.extra)
            success = self._execute_with_rest(cmdline, result)
            if success:
                break
        return success

    def _execute_with_lock(self, cmdline: List[AnyStr], result: Result) -> bool:
        @contextmanager
        def lock(cmd):
            _lock = None

            if self._exclusive:
                lock_path = Path(gettempdir(), f"{Path(cmd).name}.lock")
                # https://github.com/tox-dev/py-filelock/issues/102
                # pylint: disable=abstract-class-instantiated
                _lock = FileLock(lock_path)
                _lock.acquire(poll_intervall=10.0)

            try:
                yield _lock
            finally:
                if self._exclusive:
                    _lock.release()

        with lock(cmdline[0]):
            return self._execute_with_retry(cmdline, result)

    def _execute(self, cmdline: List[AnyStr], result: Result) -> bool:
        cmdline = self._resolve_cmdline(cmdline)
        logging.warning(" ".join(cmdline), extra=self.extra)

        success = self._execute_with_lock(cmdline, result)

        if self._test_report:
            if isinstance(self._test_report, Callable):
                self._test_report = self._test_report(result)
            result.test_report = self._test_report.apply(result)

        return success


class CommandResult(Result):
    """Full command result capsule."""
    def __init__(self, command: Command, cmdline: List):
        super().__init__(extra=command.extra)
        self.command = command
        self.cmdline = cmdline



# -*- coding: utf-8 -*-

from argparse import Namespace


class Config(Namespace):
    """Config wrapper"""

    class Any:
        """Any object"""
        def match(self, _):
            """Match"""
            return not self

    def __getattr__(self, item):
        return self.Any()



# -*- coding: utf-8 -*-

import weakref

from typing import List, Union, Type, Iterator
from enum import Enum

from ._helper import fnmatch_ex


class Axis:
    """Build matrix axis with list of configuration values.

    An axis is one dimension of a multidimensional configuration matrix.
    Each axis (dimension) can take a fixed set of values.
    """

    _instances = set()

    def __init__(self, name: str, abbrev: str = None, values: Union[List[Enum], Type[Enum]] = None, desc: str = None):
        """Initialize a new Axis.

        Args:
            name: A unique (per Builder) Axis name.
            abbrev: A unique (per Builder) Axis abbreviation.
            values: Values that can be taken on this Axis.
            desc: Descriptive text to document the intention of this Axis.
        """
        if abbrev and len(abbrev) != 1:
            raise ValueError('Parameter abbrev must be a single character string or None')

        self._name: str = name
        self._abbrev: str = abbrev
        self._values: List[Enum] = list(values) if values else []
        self._desc: str = desc

        self._instances.add(weakref.ref(self))

    def __repr__(self):
        return f'<{self.__class__.__name__} {self._name} at 0x{id(self):08X}>'

    def __getitem__(self, index: int) -> Enum:
        return self._values[index]

    @staticmethod
    def tostring(value: Enum, index: int = 0) -> str:
        """Convert Enum literal to string.

        Args:
            value: The Enum literal value.
            index: The nth tuple to be used, defaults to 0.
        """
        val = value.value
        if isinstance(val, tuple):
            return str(val[index])
        return str(val)

    @classmethod
    def get_instances(cls) -> Iterator:
        """Retrieve all class instances."""
        dead = set()
        for ref in cls._instances:
            obj = ref()
            if obj is not None:
                yield obj
            else:
                dead.add(ref)
        cls._instances -= dead

    @property
    def name(self) -> str:
        """The unique name identifying this Axis."""
        return self._name

    @property
    def abbrev(self) -> str:
        """The unique abbreviation for this Axis."""
        return self._abbrev

    @property
    def values(self) -> List[Enum]:
        """Values for this matrix Axis."""
        return self._values

    @property
    def desc(self) -> str:
        """Descriptive text to document the intention of this Axis."""
        return self._desc

    def lookup(self, pattern: str) -> List[Enum]:
        """Retrieve all Axis values matching a fnmatch pattern.

        Args:
            pattern: fnmatch pattern to match values against.

        Returns:
            A list with values matching the pattern, might be empty.
        """
        result = []
        for value in self._values:
            if fnmatch_ex(value, pattern):
                result += [value]
        return result

    def converter(self):
        """Retrieve a converter function to convert patterns to values.

        Returns:
            A function of type "pattern (string) -> List of Axis values".
        """
        def convert(pattern: str) -> List[Enum]:
            """Converter function to convert patterns to values.

            Args:
                pattern (string): A fnmatch pattern.

            Returns:
                A list of Axis value matching the given pattern, may be empty.
            """
            resolved = self.lookup(pattern)
            if resolved:
                return resolved
            raise ValueError
        return convert



# -*- coding: utf-8 -*-

import re
import shutil

from inspect import signature
from io import StringIO, TextIOBase
from typing import AnyStr, Tuple, Optional, Callable, Union

from junitparser import JUnitXml
from lxml import etree


class ReportFilter:
    """Report filter"""

    class Summary:
        """Filter summary"""

        @property
        def summary(self) -> Tuple[int, int]:
            """Report summary

            Returns: passed, executed
            """
            raise NotImplementedError

    class Result:
        """Filter result capsule."""

        def __init__(self, report, result, other):
            self._report = report
            self._result = result
            self._other = other
            self._stream = None

        def __del__(self):
            if self._stream and isinstance(self._stream, TextIOBase):
                self._stream.close()

        @staticmethod
        def cached(attr_name):
            """Decorator adding return value cache.

            Args:
                attr_name (str): The attribute to store the value.
            """
            def _cached(fn):
                def __cached(self):
                    value = getattr(self, attr_name)
                    if not value:
                        try:
                            value = fn(self)
                        except RuntimeError as e:
                            value = e
                        setattr(self, attr_name, value)
                    if isinstance(value, RuntimeError):
                        raise value
                    return value
                return __cached
            return _cached

        def getvalue(self) -> str:
            """Get the stream content as string."""
            stream = self.stream
            if hasattr(stream, 'getvalue'):
                return stream.getvalue()
            stream.seek(0)
            return stream.read()

        def write(self, filename: AnyStr):
            """Write filter input to file."""
            with open(filename, "w", encoding='utf-8') as file:
                self.stream.seek(0)
                shutil.copyfileobj(self.stream, file)

        @property
        def stream(self) -> TextIOBase:
            """Retrieve this filters text stream."""
            if self._other:
                return self._other.stream
            return StringIO()

        @property
        def result(self):
            """"Retrieve filter result string."""
            return self._result

    def __init__(self):
        self._other = None

    def __or__(self, other):
        other._other = self
        return other

    def apply(self, result):
        """Apply this filter."""
        other = None
        if self._other:
            other = self._other.apply(result)
        return self.Result(self, result, other)


class ConsoleReport(ReportFilter):
    """Load a report from console."""

    class Result(ReportFilter.Result):
        """Result capsule"""

        @property
        def stream(self) -> TextIOBase:
            if self._result:
                return self._result.output
            return StringIO()


class FileReport(ReportFilter):
    """Load a report from a file."""

    class Result(ReportFilter.Result):
        """Result capsule"""

        @property
        @ReportFilter.Result.cached("_stream")
        def stream(self) -> TextIOBase:
            try:
                return open(self._report.filename, 'r', encoding='utf-8')
            except FileNotFoundError as e:
                raise RuntimeError from e

    def __init__(self, filename):
        super().__init__()
        self.filename = filename


class CropReport(ReportFilter):
    """Crop a report from a stream."""

    class Result(ReportFilter.Result):
        """Result capsule"""

        @property
        @ReportFilter.Result.cached("_stream")
        def stream(self) -> TextIOBase:
            dump = False
            istream = self._other.stream
            ostream = StringIO()
            istream.seek(0)
            for line in istream:
                if dump:
                    ostream.write(line)
                    if re.match(self._report.last, line):
                        break
                else:
                    if re.match(self._report.first, line):
                        dump = True
                        ostream.write(line)
            return ostream

    def __init__(self, first: AnyStr, last: AnyStr):
        super().__init__()
        self.first = first
        self.last = last


class TransformReport(ReportFilter):
    """Transform an XML report using XSLT."""

    class Result(ReportFilter.Result):
        """Result capsule"""

        @property
        @ReportFilter.Result.cached("_stream")
        def stream(self) -> TextIOBase:
            try:
                ostream = StringIO()
                istream = self._other.stream
                istream.seek(0)
                dom = etree.parse(istream)
                xslt = etree.parse(self._report.xslt)
                transform = etree.XSLT(xslt)
                newdom = transform(dom)
                ostream.write(etree.tostring(newdom, pretty_print=True).decode())
                return ostream
            except etree.XMLSyntaxError as e:
                raise RuntimeError from e

    def __init__(self, xslt):
        super().__init__()
        self.xslt = xslt


class JUnitReport(ReportFilter):
    """Capture a JUnit XML report."""

    class Result(ReportFilter.Result, ReportFilter.Summary):
        """Result capsule"""

        _junit_report = None

        def _update_title(self, title):
            new_title = self._report.title
            if isinstance(new_title, Callable):
                params = [title]
                if len(signature(new_title).parameters) >= 2:
                    params += [self._result]

                return new_title(*params)
            return new_title

        @property
        @ReportFilter.Result.cached("_junit_report")
        def junit_report(self) -> JUnitXml:
            """Load a JUnit format."""
            try:
                istream = self.stream
                istream.seek(0)
                report = JUnitXml.fromfile(istream)
                if self._report.title:
                    for testsuite in report:
                        testsuite.name = self._update_title(testsuite.name)
                return report
            except etree.XMLSyntaxError as e:
                raise RuntimeError from e

        @property
        def summary(self) -> Tuple[int, int]:
            report = self.junit_report
            executed = report.tests - report.skipped
            passed = executed - report.errors - report.failures
            return passed, executed

        def write(self, filename: AnyStr):
            self.junit_report.write(filename)

    def __init__(self, title: Optional[Union[AnyStr, Callable]] = None):
        super().__init__()
        self.title = title



# -*- coding: utf-8 -*-

import json
import sys

from argparse import ArgumentParser
from enum import Enum

from .config import Config
from .runner import Runner
from .axis import Axis


class InspectRunner(Runner):
    """Inspect a matrix runner script."""

    class JSONEncoder(json.JSONEncoder):
        """Custom JSONEncoder"""
        def default(self, o):
            if isinstance(o, Axis):
                return {'abbrev': o.abbrev, 'values': o.values, 'desc': o.desc}
            if isinstance(o, Config):
                return vars(o)
            if isinstance(o, Enum):
                return str(o)
            return super().default(o)

    def __init__(self, argv=None):
        super().__init__()

        parser = ArgumentParser()
        parser.add_argument("script", type=open, help="The Matrix Runner script to inspect.")
        args, argv = parser.parse_known_args(argv)

        exec(args.script.read())  # pylint: disable=exec-used

        self.add_axis(Axis.get_instances())
        success = self.run(argv)

        if success:
            dump = json.dumps({'axes': self._axes, 'matrix': self._matrix}, cls=self.JSONEncoder)
            print(dump)

        if not success:
            sys.exit(1)
        sys.exit(0)



# -*- coding: utf-8 -*-

import weakref
from typing import Iterator, Callable

from .config import Config

FilterFunction = Callable[[Config], bool]


class Filter:
    """Matrix filter"""
    _instances = set()

    def __init__(self, fn: FilterFunction):
        self._fn = fn
        self._instances.add(weakref.ref(self))

    def __del__(self):
        self._instances.discard(weakref.ref(self))

    def __call__(self, config: Config) -> bool:
        return self._fn(config)

    @classmethod
    def get_instances(cls) -> Iterator[Callable[[Config], bool]]:
        """Retrieve all instances"""
        dead = set()
        for ref in cls._instances:
            obj = ref()
            if obj is not None:
                yield obj
            else:
                dead.add(ref)
        cls._instances -= dead

    @classmethod
    def match(cls, config: Config) -> bool:
        """Matcher"""
        return any(f(config) for f in cls.get_instances())



# -*- coding: utf-8 -*-

from enum import Enum
from typing import Callable, Optional, AnyStr, Type

# patching colorama to work on msys/cygwin terminals
# https://github.com/tartley/colorama/issues/224

try:
    # noinspection PyUnresolvedReferences
    import colorama
except ImportError:
    pass
else:
    from ._helper import is_msys_cygwin_tty

    # noinspection SpellCheckingInspection
    isatty = colorama.ansitowin32.StreamWrapper.isatty
    colorama.ansitowin32.StreamWrapper.isatty = lambda x: isatty(x) or is_msys_cygwin_tty(x)

from ._helper import partial_with_inspect
from .axis import Axis
from .action import Action, ActionFunction
from .command import Command, Result, CommandFunction
from .filter import Filter, FilterFunction
from .report import ReportFilter, ConsoleReport, CropReport, TransformReport, JUnitReport, FileReport
from .runner import Runner, RunnerApplication
from .inspect import InspectRunner

__all__ = ['Axis', 'Action', 'Command', 'Result', 'Runner', 'RunnerApplication',
           'matrix_action', 'matrix_axis', 'matrix_command', 'matrix_filter',
           'main', 'ReportFilter', 'ConsoleReport', 'CropReport',
           'TransformReport', 'JUnitReport', 'InspectRunner', 'FileReport']


def matrix_axis(name: AnyStr, abbrev: Optional[AnyStr] = None, desc: Optional[AnyStr] = None) \
        -> Callable[[Type[Enum]], Type[Enum]]:
    """Decorator to create a matrix axis from a enum.

    Args:
        name: A unique (per Builder) Axis name.
        abbrev: A unique (per Builder) Axis abbreviation.
        desc: Descriptive text to document the intention of this Axis.
    """
    def _axis(cls: Type[Enum]) -> Type[Enum]:
        doc = cls.__doc__ if not desc else desc
        setattr(cls, "axis", Axis(name, abbrev, cls, doc))
        setattr(cls, "match", lambda v, p: v in cls.axis.lookup(p))
        setattr(cls, "__getitem__", cls.axis.tostring)
        setattr(cls, "__str__", cls.axis.tostring)
        return cls
    return _axis


def matrix_action(fn: ActionFunction) -> Action:
    """Decorator for ActionFunction"""
    act = Action(fn.__name__, fn, fn.__doc__)
    setattr(fn, "action", act)
    return act


def matrix_command(**kwargs) -> Callable[[CommandFunction], Callable[..., Command]]:
    """Decorator for a CommandFunction.

    The decorated function can take arbitrary arguments and needs to return
    a list of command line arguments, see popen.

    Args:
        kwargs: Arguments passed on to Command constructor
    """
    def _command(fn: CommandFunction) -> Callable[..., Command]:
        def _command_with_args(*args, **keywords):
            return Command(partial_with_inspect(fn, *args, **keywords), **kwargs)
        return _command_with_args
    return _command


def matrix_filter(fn: FilterFunction) -> Filter:
    """Decorator for FilterFunction"""
    return Filter(fn)


main = RunnerApplication  # pylint: disable=invalid-name
